from pathlib import Path

import click

from imgtools.cli import set_log_verbosity
from imgtools.dicom import find_dicoms
from imgtools.dicom.index import DatabaseHandler, DICOMDatabaseInterface, DICOMIndexer
from imgtools.logging import logger

DEFAULT_DB_DIR = Path('.imgtools')
DEFAULT_DB_NAME = 'imgtools.db'

@click.command()
@set_log_verbosity()
@click.option(
	'--directory',
	'-d',
	type=click.Path(
		exists=True,
		file_okay=False,
		dir_okay=True,
		readable=True,
		resolve_path=True,
		path_type=Path,
	),
	help='Directory to search for DICOM files',
	required=True,
)
@click.option(
	'--update-db',
	'-u',
	is_flag=True,
	help='Force update the database',
	default=False,
)
@click.option(
	'--limit',
	'-l', 
	type=int,
	default=0,
	help='Number of files to index, 0 for all',
)
def index(directory: Path, update_db: bool = False, verbose: int = 0, quiet: bool = False, limit: int = 0 ) -> None:
	""" Index DICOM files in a directory. 
	
	For now, creates a database in the current directory at .imgtools/imgtools.db.
	
	"""
	extension = 'dcm'
	check_header = False
	logger.info('Searching for DICOM files.', args=locals())

	dicom_files = find_dicoms(
		directory=directory,
		check_header=check_header,
		recursive=True,
		extension=extension,
	)
	if not dicom_files:
		warningmsg = f'No DICOM files found in {directory}.'
		logger.warning(
			warningmsg,
			directory=directory,
			check_header=check_header,
			recursive=True,
			extension=extension,
		)
		return

	logger.info('DICOM find successful.', count=len(dicom_files))

	db_path = DEFAULT_DB_DIR / DEFAULT_DB_NAME

	db_path.parent.mkdir(parents=True, exist_ok=True)

	db_handler = DatabaseHandler(db_path=db_path, force_delete=update_db)

	indexer = DICOMIndexer(db_handler=db_handler)

	if limit:
		dicom_files = dicom_files[:limit]

	logger.debug('Building index.', count=len(dicom_files), limit=limit)
	# Build index
	indexer.build_index_from_files(dicom_files)	
	
	_ = DICOMDatabaseInterface(db_handler=db_handler)

	logger.info('Indexing complete.')

	click.echo(f'Indexed {len(dicom_files)} files to {db_path}')

if __name__ == '__main__':
	index()




from __future__ import annotations

from collections import Counter
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, List, Optional

from pydicom import dcmread
from sqlalchemy import JSON, Column, ForeignKey, String, Table
from sqlalchemy.orm import relationship
from sqlalchemy.orm.decl_api import registry

from imgtools.logging import logger

if TYPE_CHECKING:
	from pathlib import Path

mapper_registry = registry()


def repr_mixin(cls) -> type:  # type: ignore  # noqa: ANN001
	"""
	A decorator to add a customizable __repr__ method to SQLAlchemy ORM models.

	Includes scalar fields (e.g., columns) and prints the length of lists for relationships.
	"""

	def __repr_method__(self) -> str:  # type: ignore  # noqa: N807, ANN001
		# Fetch attributes and handle lists separately
		attrs = {
			key: (len(value) if isinstance(value, list) else value)
			for key, value in vars(self).items()
			if not key.startswith('_') and isinstance(value, (str, int, float, list))
		}
		class_name = self.__class__.__name__
		fields = ',\n  '.join(f'{key}={value}' for key, value in attrs.items())
		return f'<{class_name}(\n  {fields}\n)>'

	cls.__repr__ = __repr_method__
	return cls

####################################################################################################
# PATIENT
####################################################################################################

# Patient Table
@repr_mixin
@mapper_registry.mapped
@dataclass
class Patient:
	__table__ = Table(
		'patients',
		mapper_registry.metadata,
		Column('PatientID', String, primary_key=True, index=True),  # PRIMARY KEY
	)

	# Metadata from DICOM
	PatientID: str = field(init=True)

	# Relationships
	studies: List[Study] = field(default_factory=list)

	__mapper_args__ = {  # type: ignore
		'properties': {
			'studies': relationship('Study', back_populates='patient'),
		}
	}

	@classmethod
	def from_metadata(cls, metadata: Dict[str, str]) -> Patient:
		"""
		Create a Patient instance from metadata.

		Parameters
		----------
		metadata : Dict[str, Optional[str]]
				Patient metadata.

		Returns
		-------
		Patient
				A Patient instance.
		"""
		return cls(
			PatientID=metadata['PatientID'],
		)

	@property
	def num_studies(self) -> int:
		"""
		Get the number of studies for the patient.

		Returns
		-------
		int
				The number of studies.
		"""
		return len(self.studies)

	@property
	def num_series(self) -> int:
		"""
		Get the number of series for the patient.

		Returns
		-------
		int
				The number of series.
		"""
		return sum(len(study.series) for study in self.studies)

	@property
	def modalities(self) -> Dict[str, int]:
		"""
		Get the list of modalities for the patient.

		Returns
		-------
		List[str]
				The list of modalities.
		"""
		modalities = [modality for study in self.studies for modality in study.modalities]
		return dict(Counter(modalities))

	@property
	def series(self) -> List[Series]:
		"""
		Get the series for the patient.

		Returns
		-------
		List[Series]
				The series.
		"""
		return [series for study in self.studies for series in study.series]

	@property
	def rtstructs(self) -> List[Series]:
		"""
		Get the RTSTRUCT series for the patient.

		Returns
		-------
		List[Series]
				The RTSTRUCT series.
		"""
		return [series for series in self.series if series.Modality == 'RTSTRUCT']


####################################################################################################
# STUDY
####################################################################################################

# Study Table
@repr_mixin
@mapper_registry.mapped
@dataclass
class Study:
	__table__ = Table(
		'studies',
		mapper_registry.metadata,
		Column('StudyInstanceUID', String, primary_key=True, index=True),  # PRIMARY KEY
		Column('PatientID', String, ForeignKey('patients.PatientID'), nullable=False),
	)

	# Metadata from DICOM
	StudyInstanceUID: str = field(init=True)  # PRIMARY KEY
	PatientID: str = field(init=True)  # FOREIGN KEY

	# Relationships
	patient: Patient = field(init=False)
	series: List[Series] = field(default_factory=list)

	__mapper_args__ = {  # type: ignore
		'properties': {
			'patient': relationship('Patient', back_populates='studies'),
			'series': relationship('Series', back_populates='study'),
		}
	}

	@classmethod
	def from_metadata(cls, metadata: Dict[str, str]) -> Study:
		"""
		Create a Study instance from metadata.

		Parameters
		----------
		metadata : Dict[str, Optional[str]]
				Study metadata.

		Returns
		-------
		Study
				A Study instance.
		"""
		return cls(
			StudyInstanceUID=metadata['StudyInstanceUID'],
			PatientID=metadata['PatientID'],
		)

	@property
	def num_files(self) -> int:
		"""
		Get the number of files in the study.

		Returns
		-------
		int
				The number of files.
		"""
		return sum(series.num_files for series in self.series)

	@property
	def modalities(self) -> Dict[str, int]:
		"""
		Get the list of modalities in the study.

		Returns
		-------
		List[str]
				The list of modalities.
		"""

		modalities = [series.Modality for series in self.series]
		return dict(Counter(modalities))

	@property
	def unique_modalities(self) -> List[str]:
		"""
		Get the unique modalities in the study.

		Returns
		-------
		List[str]
				The unique modalities.
		"""
		return list(self.modalities.keys())


####################################################################################################
# SERIES, RTSTRUCTSERIES, ...
####################################################################################################


@repr_mixin
@mapper_registry.mapped
@dataclass
class Series:
	__table__ = Table(
		'series',
		mapper_registry.metadata,
		Column('SeriesInstanceUID', String, primary_key=True, index=True),  # PRIMARY KEY
		Column('StudyInstanceUID', String, ForeignKey('studies.StudyInstanceUID'), nullable=False),
		Column('Modality', String, nullable=False),
		Column('_RTSTRUCT', JSON, nullable=True),
		Column('_MR', JSON, nullable=True),
	)

	# Metadata from DICOM
	SeriesInstanceUID: str = field(init=True)  # PRIMARY KEY
	StudyInstanceUID: str = field(init=True)  # FOREIGN KEY
	Modality: str = field(init=True)
	_RTSTRUCT: Optional[Dict[str, str]] = field(default_factory=dict)
	_MR: Optional[Dict[str, str]] = field(default_factory=dict)

	# Relationships
	study: Study = field(init=False)
	images: List[Image] = field(default_factory=list)

	__mapper_args__ = {  # type: ignore
		'properties': {
			'study': relationship('Study', back_populates='series'),
			'images': relationship('Image', back_populates='series'),
		},
	}

	@classmethod
	def from_metadata(cls, metadata: Dict[str, str], file_path: Path) -> Series:

		base_cls = cls(
			SeriesInstanceUID=metadata['SeriesInstanceUID'],
			StudyInstanceUID=metadata['StudyInstanceUID'],
			Modality=metadata['Modality'],
		)

		match metadata['Modality']:
			case 'RTSTRUCT':
				try:
					ds = dcmread(file_path, stop_before_pixels=True)
					logger.info(f'Reading RTSTRUCT file: {file_path}')
					roi_names = [roi.ROIName for roi in ds.StructureSetROISequence]
					rfrs = ds.ReferencedFrameOfReferenceSequence[0]
					ref_series_seq = rfrs.RTReferencedStudySequence[0]\
						.RTReferencedSeriesSequence[0]\
						.SeriesInstanceUID
					frame_of_reference = rfrs.FrameOfReferenceUID
					base_cls._RTSTRUCT = {
						'ROINames': ",".join(roi_names),
						'RTReferencedSeriesUID': ref_series_seq,
						'FrameOfReferenceUID': frame_of_reference,
					}
				except Exception as e:
					logger.exception(f'Error reading RTSTRUCT file: {file_path}', error=str(e))
					raise e
			case 'MR':
				try:
					ds = dcmread(file_path, stop_before_pixels=True)
					logger.info(f'Reading MR file: {file_path}')
					_mr = {
						'RepetitionTime': ds.RepetitionTime,
						'EchoTime': ds.EchoTime,
						'SliceThickness': ds.SliceThickness,
						'ScanningSequence': ds.ScanningSequence,
						'MagneticFieldStrength': ds.MagneticFieldStrength,
						'ImagedNucleus': ds.ImagedNucleus,
					}

					base_cls._MR = {k: str(v) for k, v in _mr.items()}
				except Exception as e:
					logger.exception(f'Error reading MR file: {file_path}', error=str(e))
					raise e
			case _:
				pass

		return base_cls

	@property
	def num_files(self) -> int:
		return len(self.images)

	@property
	def RTReferencedSeriesUID(self) -> Optional[str]:  # noqa: N802
		return self._RTSTRUCT.get('RTReferencedSeriesUID', None) if self._RTSTRUCT else None

	@property
	def ROINames(self) -> Optional[str]: # noqa: N802
		return self._RTSTRUCT.get('ROINames', None) if self._RTSTRUCT else None




####################################################################################################
# IMAGE
####################################################################################################

# Image Table
@repr_mixin
@mapper_registry.mapped
@dataclass
class Image:
	__table__ = Table(
		'images',
		mapper_registry.metadata,
		Column('FilePath', String, primary_key=True, index=True),  # PRIMARY KEY
		Column('SOPInstanceUID', String, nullable=False),
		Column('SeriesInstanceUID', String, ForeignKey('series.SeriesInstanceUID'), nullable=False),
	)

	# Metadata from DICOM
	FilePath: str = field(init=True)
	SOPInstanceUID: str = field(init=True)
	SeriesInstanceUID: str = field(init=True)

	# Relationships
	series: Series = field(init=False)

	__mapper_args__ = {  # type: ignore
		'properties': {
			'series': relationship('Series', back_populates='images'),
		}
	}

	@classmethod
	def from_metadata(cls, metadata: Dict[str, str], file_path: Path) -> Image:
		"""
		Create a Image instance from metadata.

		Parameters
		----------
		metadata : Dict[str, Optional[str]]
				Image metadata.

		Returns
		-------
		Image
				A Image instance.
		"""
		return cls(
			SOPInstanceUID=metadata['SOPInstanceUID'],
			SeriesInstanceUID=metadata['SeriesInstanceUID'],
			FilePath=file_path.as_posix(),
		)



from contextlib import contextmanager
from pathlib import Path
from typing import Generator

from sqlalchemy import create_engine
from sqlalchemy.orm import Session, sessionmaker

from imgtools.dicom.index.models import mapper_registry
from imgtools.logging import logger


class DatabaseHandler:
	"""
	Manages database operations using SQLAlchemy ORM.
	"""

	def __init__(self, db_path: Path, force_delete: bool = False) -> None:
		"""
		Initialize the DatabaseHandler with an SQLite database.

		Parameters
		----------
		db_path : Path
			Path to the SQLite database file.
		force_delete : bool
			Whether to delete the existing database file if it exists.
		"""
		if force_delete and db_path.exists():
			logger.debug('Deleting existing database file.', db_path=db_path)
			db_path.unlink()  # Delete the existing database file

		logger.debug('Creating database engine.', db_path=db_path)
		self.engine = create_engine(f'sqlite:///{db_path}', future=True)
		mapper_registry.metadata.create_all(self.engine)  # Create tables
		self.Session = sessionmaker(bind=self.engine)

	@contextmanager
	def session(self) -> Generator[Session, None, None]:
		"""
		Create a new SQLAlchemy session.

		Yields
		------
		Session
			A SQLAlchemy session object.
		"""
		session = self.Session()
		try:
			yield session
			session.commit()  # Commit the transaction
		except Exception:
			session.rollback()  # Rollback on exception
			raise
		finally:
			session.close()
